from openalex_analysis.data.entities_data import config
from openalex_analysis.data.entities_data import load_config_from_file
from openalex_analysis.data.entities_data import log_oa

from openalex_analysis.data.entities_data import EntitiesData
from openalex_analysis.data.entities_data import WorksData
from openalex_analysis.data.entities_data import AuthorsData
from openalex_analysis.data.entities_data import SourcesData
from openalex_analysis.data.entities_data import InstitutionsData
from openalex_analysis.data.entities_data import TopicsData
from openalex_analysis.data.entities_data import ConceptsData
from openalex_analysis.data.entities_data import PublishersData

from openalex_analysis.data.entities_data import get_entity_type_from_id
from openalex_analysis.data.entities_data import get_name_of_entity
from openalex_analysis.data.entities_data import get_info_about_entity
from openalex_analysis.data.entities_data import check_if_entity_exists


__all__ = [
    "config",
    "load_config_from_file",
    "log_oa",
    "EntitiesData",
    "WorksData",
    "AuthorsData",
    "SourcesData",
    "InstitutionsData",
    "TopicsData",
    "ConceptsData",
    "PublishersData",
    "get_entity_type_from_id",
    "get_name_of_entity",
    "get_info_about_entity",
    "check_if_entity_exists",
]



import sys
from os.path import isdir
import shutil
import numpy as np

sys.path.append("..")

from openalex_analysis.analysis import config, load_config_from_file
from openalex_analysis.data import WorksData
from openalex_analysis.analysis import WorksAnalysis, InstitutionsAnalysis, AuthorsAnalysis
from openalex_analysis.plot import WorksPlot

# set the default configuration (this avoids using the configuration defined in the file
# ~/openalex-analysis/openalex-analysis-conf.toml)
from openalex_analysis.data.entities_data import set_default_config

set_default_config()

load_config_from_file("openalex-analysis-conf.toml")

institution_src_id = "I138595864"

regime_shift_topic_id = "T13377"

# Use a specific folder for the tests to be able to clear the cache
config.project_data_folder_path = "./data"

# remove data (cache) that may have been downloaded in the previous test
if isdir(config.project_data_folder_path):
    shutil.rmtree(config.project_data_folder_path)


def test_download_dataset_1():
    # basic test with works
    wa = WorksAnalysis(institution_src_id)
    first_author = wa.entities_df.at[0, "authorships"][0]

    assert isinstance(first_author['author']['display_name'], str)


def test_download_dataset_2():
    # test with works and extra filter
    extra_filters = {
        'publication_year': 2020,
        'authorships': {'institutions': {'id': institution_src_id}},
    }

    wa = WorksAnalysis(extra_filters=extra_filters)
    first_author = wa.entities_df.at[0, "authorships"][0]

    assert isinstance(first_author['author']['display_name'], str)


def test_download_dataset_3():
    # test with a topic dataset
    wa = WorksPlot(regime_shift_topic_id)
    print(wa.entities_df.authorships[0])
    first_author = wa.entities_df.at[0, "authorships"][0]

    assert isinstance(first_author['author']['display_name'], str)


def test_download_dataset_4():
    # basic test with authors
    aa = AuthorsAnalysis(institution_src_id)
    print(aa.entities_df.columns)
    first_author_name = aa.entities_df.at[0, "display_name"]

    assert isinstance(first_author_name, str)


def test_analysis_1():
    wa = WorksAnalysis(institution_src_id)

    wa.create_element_used_count_array('reference')

    assert isinstance(wa.element_count_df.iloc[0, 0], np.int64)


def test_get_multiple_entities_from_id_works():
    # test with a list of 3 articles
    entities_ids = [
        "W1999167944",
        "W2096885696",
        "W2126902408",
    ]
    entities_names = [
        "Planetary boundaries: Guiding human development on a changing planet",
        "A safe operating space for humanity",
        "Solutions for a cultivated planet",
    ]
    res = WorksAnalysis().get_multiple_entities_from_id(entities_ids, return_dataframe=False)
    for i in range(len(entities_names)):
        assert entities_names[i] == res[i]["display_name"]
    # TODO: add a test with more than 100 works


def test_get_multiple_entities_from_id_institutions():
    # test with a list of 3 institutions
    entities_ids = [
        "I138595864",
        "I140494188",
        "I000000000"
    ]
    entities_names = [
        "Stockholm Resilience Centre",
        "Université de Technologie de Troyes",
        None
    ]
    res = InstitutionsAnalysis().get_multiple_entities_from_id(entities_ids, return_dataframe=False)
    for i in range(len(entities_names)):
        if entities_ids[i] == "I000000000":
            assert res[i] is None
        else:
            assert entities_names[i] == res[i]["display_name"]
    # TODO: add a test with more than 100 institutions


def test_get_multiple_entities_from_id_works_to_df():
    # test with a list of 3 articles
    entities_ids = [
        "W1999167944",
        "W2096885696",
        "W2126902408",
    ]
    entities_names = [
        "Planetary boundaries: Guiding human development on a changing planet",
        "A safe operating space for humanity",
        "Solutions for a cultivated planet",
    ]
    res = WorksAnalysis().get_multiple_entities_from_id(entities_ids)
    for i in range(len(entities_names)):
        assert entities_names[i] == res.at[i, "display_name"]
    assert "abstract" in res.columns


def test_get_multiple_works_from_doi():
    # test with a list of 3 articles
    article_dois = [
        "https://doi.org/10.1038/461472a",
        # added upper case to check that the ordering is working with uppercases:
        "https://doi.org/10.1126/SCIENCE.1259855",
        "https://doi.org/10.1038/nature10452",
    ]
    article_names = [
        "A safe operating space for humanity",
        "Planetary boundaries: Guiding human development on a changing planet",
        "Solutions for a cultivated planet",
    ]
    res = WorksData().get_multiple_works_from_doi(article_dois, return_dataframe=False)
    for i in range(len(article_dois)):
        assert article_names[i] == res[i]["display_name"]
    # TODO: add a test with more than 60 works


def test_concept_yearly_count():
    concept_sustainability_id = 'C66204764'
    # create the filter for the API to get only the articles about sustainability
    sustainability_concept_filter = {"concepts": {"id": concept_sustainability_id}}

    # set the years we want to count
    count_years = list(range(2004, 2024))

    institution_ids_list = ["I138595864", "I140494188"]
    # ["Stockholm Resilience Centre", "University of Technology of Troyes"]

    # create a list of dictionaries with each dictionary containing the ID, name and filter for each institution
    entities_ref_to_count = [{}] * len(institution_ids_list)
    for i in range(len(institution_ids_list)):
        entities_ref_to_count[i] = {'entity_from_id': institution_ids_list[i],
                                    'extra_filters': sustainability_concept_filter}

    wplt = WorksPlot()
    wplt.create_element_used_count_array('concept', entities_ref_to_count, count_years=count_years)

    wplt.add_statistics_to_element_count_array(sort_by='sum_all_entities')

    wplt.get_figure_time_series_element_used_by_entities().write_image("Plot_yearly_usage_sustainability_SRC_UTT.svg",
                                                                       width=900, height=350)


def test_generating_collaboration_map():
    wplt = WorksPlot("I138595864")

    wplt.get_collaborations_with_institutions()

    wplt.get_figure_collaborations_with_institutions()



# Romain THOMAS 2024

import os
from os.path import exists, join, isdir, isfile, expanduser
import psutil
from pathlib import Path
import hashlib  # to generate file names
from time import time
import logging
import tomllib

import pyalex.api
from tqdm import tqdm
import pandas as pd
import requests

from pyalex import Works, Authors, Sources, Institutions, Topics, Concepts, Publishers, config

# define a custom logging
log_oa = logging.getLogger(__name__)
log_oa.addHandler(logging.StreamHandler())
# log_oa.addHandler(logging.FileHandler(__name__ + ".log"))


class AnalysisConfig(dict):
    """
    OpenAlex Analysis configuration class. This class contains the settings used by openalex-analysis (some settings
    are passed to pyalex).

    Tu use it, import the class and set the parameters as follows:
    
    .. code-block:: python

        from openalex_analysis.plot import config

        config.n_max_entities = 10000

    * **email** (*str*) - Your Email for the OpenAlex API. Allows you to use the polite pool (see OpenAlex documentation). The default value is None (not using the polite pool).
    * **api_key** (*str*) - Your OpenAlex API key, if you have one. The default value is None.
    * **openalex_url** (*str*) - OpenAlex API URL or your self-hosted API URL. The default value is "https://api.openalex.org".
    * **http_retry_times** (*int*) - maximum number of retries when querying the OpenAlex API in HTTP. The default value is 3.
    * **disable_tqdm_loading_bar** (*bool*) - To disable the tqdm loading bar. The default is False.
    * **n_max_entities** (*int*) - Maximum number of entities to download (the default value is to download maximum 10 000 entities). If set to None, no limitation will be applied.
    * **project_data_folder_path** (*str*) - Path to the folder containing the data downloaded from the OpenAlex API (these data are stored in compressed parquet files and used as a cache). The default path is "~/openalex-analysis/data".
    * **parquet_compression** (*str*) - Type of compression for the parquet files used as cache (see the Pandas documentation). The default value is "brotli".
    * **max_storage_percent** (*int*) - When the disk capacity reaches this percentage, cached parquet files will be deleted. The default value is 95.
    * **max_storage_files** (*int*) - When the cache folder reaches this number of files, cached parquet files will be deleted. The default value is 10000.
    * **max_storage_size** (*int*) - When the cache folder reached this size (in bytes), cached parquet files will be deleted. The default value is 5e9 (5 GB).
    * **min_storage_files** (*int*) - Before deleting files, we check if we exceed the minimum number of files and folder size. If one of those minimum if exceeded, we allow the program to delete cached parquet files. This is to avoid the setting max_storage_percent to delete every cached file when the disk is almost full. The default value is 1000.
    * **min_storage_size** (*int*) - Before deleting files, we check if we exceed the minimum number of files and folder size. If one of those minimum if exceeded, we allow the program to delete cached parquet files. This is to avoid the setting max_storage_percent to delete every cached file when the disk is almost full. The default value is 5e8 (500 MB).
    * **cache_max_age** (*int*) - Maximum age of the cache in days. The default value is 365.
    * **log_level** (*str*) - The log detail level for openalex-analysis (library specific). The log_level must be 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'. The default value 'WARNING'.
    """
    def __getattr__(self, key):
        return super().__getitem__(key)


    def __setattr__(self, key, value):
        if key == "log_level":
            match value:
                case 'DEBUG':
                    log_oa.setLevel(logging.DEBUG)
                case 'INFO':
                    log_oa.setLevel(logging.INFO)
                case 'WARNING':
                    log_oa.setLevel(logging.WARNING)
                case 'ERROR':
                    log_oa.setLevel(logging.ERROR)
                case 'CRITICAL':
                    log_oa.setLevel(logging.CRITICAL)
                case _:
                    raise ValueError("The log_level must be 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'")

        return super().__setitem__(key, value)


config = AnalysisConfig()


def set_default_config():
    """
    Set the default configuration of the library. This function is called is no configuration file is found.
    """
    log_oa.info(f"Setting the default configuration")
    config.email = None
    config.api_key = None
    config.openalex_url = "https://api.openalex.org"
    config.http_retry_times = 3
    config.disable_tqdm_loading_bar = False
    config.n_max_entities = 10000
    config.project_data_folder_path = join(expanduser("~"), "openalex-analysis", "data")
    config.parquet_compression = "brotli"
    config.max_storage_percent = 95
    config.max_storage_files = 10000
    config.max_storage_size = 5e9
    config.min_storage_files = 1000
    config.min_storage_size = 5e8
    config.cache_max_age = 365
    config.log_level = 'WARNING'


def load_config_from_file(config_path: str):
    """
    Load and set the configration of the library from a .toml file. When the library is imported, if a configuration
    file exists at "~/openalex-analysis/openalex-analysis-conf.toml", it is automatically loaded.

    :param config_path: The path of the configuration file.
    :type config_path: str
    """
    log_oa.info(f"Loading the configuration from the file {config_path}")
    with open(config_path, "rb") as f:
        config_data = tomllib.load(f)
    if "email" in config_data.keys():
        config.email = config_data['email']
    if "api_key" in config_data.keys():
        config.api_key = config_data['api_key']
    if "openalex_url" in config_data.keys():
        config.openalex_url = config_data['openalex_url']
    if "http_retry_times" in config_data.keys():
        config.http_retry_times = config_data['http_retry_times']
    if "disable_tqdm_loading_bar" in config_data.keys():
        config.disable_tqdm_loading_bar = config_data['disable_tqdm_loading_bar']
    if "n_max_entities" in config_data.keys():
        config.n_max_entities = config_data['n_max_entities']
    if "project_data_folder_path" in config_data.keys():
        config.project_data_folder_path = config_data['project_data_folder_path']
    if "parquet_compression" in config_data.keys():
        config.parquet_compression = config_data['parquet_compression']
    if "max_storage_percent" in config_data.keys():
        config.max_storage_percent = config_data['max_storage_percent']
    if "max_storage_files" in config_data.keys():
        config.max_storage_files = config_data['max_storage_files']
    if "max_storage_size" in config_data.keys():
        config.max_storage_size = config_data['max_storage_size']
    if "min_storage_files" in config_data.keys():
        config.min_storage_files = config_data['min_storage_files']
    if "min_storage_size" in config_data.keys():
        config.min_storage_size = config_data['min_storage_size']
    if "cache_max_age" in config_data.keys():
        config.cache_max_age = config_data['cache_max_age']
    if "log_level" in config_data.keys():
        config.log_level = config_data['log_level']


# if the config file exist, load the configuration from it otherwise load the default configuration
if isfile(join(expanduser("~"), "openalex-analysis","openalex-analysis-conf.toml")):
    load_config_from_file(join(expanduser("~"), "openalex-analysis", "openalex-analysis-conf.toml"))
else:
    set_default_config()


class EntitiesData:
    """
    This class contains methods to download data from the OpenAlex API and manage + cache those datasets locally
    """

    def __init__(self,
                 entity_from_id: str | None = None,
                 extra_filters: dict | None = None,
                 database_file_path: dict | None = None,
                 create_dataframe: bool = True,
                 load_only_columns: list[str] | None = None,
                 ):
        """

        :param entity_from_id: The entity identifier (eg an institution id) from which to take the entities (eg the works of this institution) to analyse. If not provided, the default value is None and the entities will be downloaded bases on the extra_filters value.
        :type entity_from_id: str | None
        :param extra_filters: Optional filters, refer to the documentation of openalex and pyalex for the format. The default value is None.
        :type extra_filters: dict | None
        :param database_file_path: The database file (parquet or csv) path to force the analysis over datas in a specific file. The default value is None to use the data from the OpenAlex API or the cached data.
        :type database_file_path: str | None
        :param create_dataframe: Create the dataframe at the initialisation (and download the data if allowed and entitie_from_id or extra_filters is provided). The default value is True.
        :type create_dataframe: bool
        :param load_only_columns: Load only the specified columns from the parquet file. Everything will be downloaded anyway. The default value is None.
        :type load_only_columns: str | None
        """
        self.per_page = 200  # maximum allowed by the API

        self.entity_from_id = entity_from_id
        self.entity_from_type = None
        self.extra_filters = extra_filters
        self.database_file_path = database_file_path
        self.load_only_columns = load_only_columns

        # a dataframe containing entities related to the instance
        self.entities_df = None

        # # Dataframe with the entities filtered with multi concepts
        # self.entities_multi_filtered_df = None

        # Dataframe with all the element (count_element_type) of the entity(ies) and their count + the count for other
        # entities (optional). The creation needs to be manually started.
        self.element_count_df = pd.DataFrame()

        # to display a loading bar on the web interface
        self.entity_downloading_progress_percentage = 0
        self.create_element_count_array_progress_percentage = 0
        self.create_element_count_array_progress_text = ""
        self.count_element_type = None
        self.count_element_years = None
        self.count_entities_cols = []

        # initialize the values only if entity_from_type is known
        if self.entity_from_id is not None or self.extra_filters is not None:
            if self.entity_from_id is not None:
                self.entity_from_type = self.get_entity_type_from_id(self.entity_from_id)
            if self.database_file_path is None:
                self.database_file_path = join(config.project_data_folder_path, self.get_database_file_name())
            if create_dataframe:
                self.load_entities_dataframe()

    def get_count_entities_matched(self, query_filters: dict) -> int:
        """
        Gets and return the number of entities which match the query filters.

        :param query_filters: The query filters.
        :type query_filters: dict
        :return: The entities count matched.
        :rtype: int
        """
        results, meta = self.EntityOpenAlex().filter(**query_filters).get(per_page=1, return_meta=True)
        return meta['count']

    def get_api_query(self) -> dict:
        """
        Gets the api query from the parameters of the instance.

        :return: The api query
        :rtype: dict
        """
        if self.entity_from_id is not None:
            query_filters = {self.get_entity_type_string_name(self.entity_from_type): {"id": self.entity_from_id}}
            # special cases:
            # concept of institution as the query key word is x_concepts instead of concepts
            if self.get_entity_type_from_id() == Institutions and self.get_entity_type_from_id(
                    self.entity_from_id) == Concepts:
                query_filters = {'x_concepts': {"id": self.entity_from_id}}
            if self.get_entity_type_from_id(self.entity_from_id) == Authors:
                query_filters = {'author': {"id": self.entity_from_id}}
            # when querying authors from an institution, we use the attribute affiliations.institution.id because at the
            # time of the implementation (05/09/2024), the API filter last_known_institutions.id was not working
            if self.EntityOpenAlex == Authors and self.get_entity_type_from_id() == Institutions:
                query_filters = {'affiliations': {'institution': {"id": self.entity_from_id}}}
        else:
            query_filters = {}
        if self.extra_filters is not None:
            query_filters = query_filters | self.extra_filters
        log_oa.info(f"query: {query_filters}")
        return query_filters


    def convert_entities_list_to_df(self, entities_list: list) -> pd.DataFrame:
        """
        Converts a list of entities (a list of PyAlex objects) downloaded from the OpenAlex API into a dataframe.
        :param entities_list: The list of entities (PyAlex objects).
        :param entities_list: list
        :return: The entities in a DataFrame.
        :rtype: pd.DataFrame
        """
        entities_list_df = pd.DataFrame.from_records(entities_list)
        if self.EntityOpenAlex == Works:
            entities_list_df = entities_list_df.rename(columns={'extracted_abstract': 'abstract'})
        return entities_list_df


    def filter_and_format_entity_data_from_api_response(self, entity: dict):
        """
        Filter and format the data downloaded from the API.
        This is a placeholder as not all entity types have this function defined.
        """
        pass


    def download_list_entities(self):
        """
        Downloads the entities which match the parameters of the instance, and store the dataset as a parquet file .
        """
        log_oa.info(f"Downloading list of {self.get_entity_type_string_name()}")
        if self.entity_from_id is not None:
            log_oa.info(f"of the {self.get_entity_type_string_name(self.entity_from_type)[0:-1]} {self.entity_from_id}")
        if self.extra_filters is not None:
            log_oa.info(f"with extra filters: {self.extra_filters}")

        query = self.get_api_query()
        log_oa.info(f"Query to download from the API: {query}")

        count_entities_matched = self.get_count_entities_matched(query)

        if config.n_max_entities is None or config.n_max_entities > count_entities_matched:
            n_entities_to_download = count_entities_matched
            print(f"All the {n_entities_to_download} entities will be downloaded")
            log_oa.info(f"All the {n_entities_to_download} entities will be downloaded")
        else:
            n_entities_to_download = config.n_max_entities
            print(f"Only {n_entities_to_download} entities will be downloaded (out of {count_entities_matched})")
            log_oa.info(f"Only {n_entities_to_download} entities will be downloaded (out of {count_entities_matched})")

        # create a list to store the entities
        entities_list = [None] * n_entities_to_download

        # create the pager entity to iterate over the pages of entities to download
        pager = self.EntityOpenAlex().filter(**query).paginate(per_page=self.per_page, n_max=n_entities_to_download)

        log_oa.info("Downloading the list of entities thought the OpenAlex API...")
        with tqdm(total=n_entities_to_download, disable=config.disable_tqdm_loading_bar) as pbar:
            i = 0
            self.entity_downloading_progress_percentage = 0
            for page in pager:
                # add the downloaded entities in the main list
                for entity in page:
                    self.filter_and_format_entity_data_from_api_response(entity)
                    if i < n_entities_to_download:
                        entities_list[i] = entity
                    else:
                        # useless ?
                        entities_list.append(entity)
                        log_oa.warning("entities_list was too short, appending the entity")
                    i += 1
                # update the progress bar
                pbar.update(self.per_page)
                # update the progress percentage variable
                self.entity_downloading_progress_percentage = i / n_entities_to_download * 100 if i else 0
        self.entity_downloading_progress_percentage = 100

        log_oa.info("Converting the entities list downloaded to a DataFrame...")
        entities_list_df = self.convert_entities_list_to_df(entities_list)

        if not isdir(config.project_data_folder_path):
            log_oa.info("Creating the directory to store the data from OpenAlex")
            os.makedirs(config.project_data_folder_path)
        log_oa.info("Checking space left on disk...")
        self.auto_remove_databases_saved()
        # save as compressed parquet file
        log_oa.info("Saving the list of entities as a parquet file...")
        entities_list_df.to_parquet(self.database_file_path, compression=config.parquet_compression)

    def load_entities_dataframe(self):
        """
        Loads an entities dataset from file (or download it if needed and allowed by the instance) to the dataframe of
        the instance.
        """
        log_oa.info(f"Loading dataframe of {self.get_entity_type_string_name()}")
        if self.entity_from_id is not None:
            log_oa.info(f"of the {self.get_entity_type_string_name(self.entity_from_type)[0:-1]} {self.entity_from_id}")
        if self.extra_filters is not None:
            log_oa.info(f"with extra filters: {self.extra_filters}")

        # # check if the database file exists
        if not exists(self.database_file_path):
            self.download_list_entities()
        # check the age of the cache (aka last date of modification)
        else:
            age_in_days = (time() - os.stat(self.database_file_path).st_mtime) / 86400
            if age_in_days > config.cache_max_age:
                os.remove(self.database_file_path)
                log_oa.info(f"Removed file {self.database_file_path} (age (days): {int(age_in_days)})")
                self.download_list_entities()
        log_oa.info("Loading the list of entities from a parquet file...")
        try:
            self.entities_df = pd.read_parquet(self.database_file_path, columns=self.load_only_columns)
        except:
            # TODO: better manage the exception
            # couldn't load the parquet file (eg no row in parquet file so error because can't find columns to load)
            self.entities_df = pd.DataFrame()

    def auto_remove_databases_saved(self):
        """
        Remove databases files (the cached data downloaded from OpenAlex) if the storage is full, if there are too many
        files or if the cache uses too much space. It keeps the last accessed files with a minimum of files number, and
        folder size.
        """
        def max_cache_storage_usage_reached():
            nb_files = len(os.listdir(config.project_data_folder_path))
            size = sum(f.stat().st_size for f in Path(config.project_data_folder_path).glob('**/*') if f.is_file())
            # if we are reaching the threshold to delete the cache (this is useful if the program is running on a system
            # with a nearly full disk to avoid the limit from config.max_storage_percent to delete every cached file),
            # we check that one of the minimum of files number of folder size is exceeded:
            if nb_files > config.min_storage_files or size > config.min_storage_size:
                if psutil.disk_usage(config.project_data_folder_path).percent > config.max_storage_percent:
                    return True
                if nb_files > config.max_storage_files:
                    return True
                if size > config.max_storage_size:
                    return True
            return False


        while max_cache_storage_usage_reached():
            first_accessed_file = None
            first_accessed_file_time = 0
            for file in os.listdir(config.project_data_folder_path):
                if file.endswith(".parquet"):
                    if os.stat(join(config.project_data_folder_path,
                                    file)).st_atime < first_accessed_file_time or first_accessed_file_time == 0:
                        first_accessed_file_time = os.stat(join(config.project_data_folder_path, file)).st_atime
                        first_accessed_file = file
            if first_accessed_file is None:
                log_oa.warning("No more file to delete.")
                log_oa.warning(f"Space used on disk: {psutil.disk_usage(config.project_data_folder_path).percent} %")
                break
            os.remove(join(config.project_data_folder_path, first_accessed_file))
            log_oa.info(f"Removed file {join(config.project_data_folder_path, first_accessed_file)} "
                        f"(last used: {first_accessed_file_time})")


    def get_database_file_name(self,
                               entity_from_id: str | None = None,
                               entity_type: pyalex.api.BaseOpenAlex | None = None,
                               db_format: str = "parquet"
                               ) -> str:
        """
        Gets the database file name according to the parameters of the object or the arguments given.

        :param entity_from_id: The instance entity identifier (eg a concept id) which was used to filter the entities (eg works) in the database. If nothing is provided, the instance entity id will be used. Default is None.
        :type entity_from_id: str | None
        :param entity_type: The entity type in the database (eg works). If nothing is provided, the instance entity id will be used. Default is None.
        :type entity_type: pyalex.api.BaseOpenAlex | None
        :param db_format: The database file format. The default is "parquet".
        :type db_format: str
        :return: The database file name
        :rtype: str
        """
        if entity_from_id is None:
            entity_from_id = self.entity_from_id
        if entity_type is None:
            entity_type = self.EntityOpenAlex
        file_name = self.get_entity_type_string_name(entity_type)
        if entity_from_id is not None:
            file_name += "_from_" + entity_from_id
        if self.extra_filters is not None:
            file_name += "_" + str(self.extra_filters).replace("'", '').replace(":", '').replace(' ', '_')
        # add the data format version, we increment this number if the format (e.g. columns) change in the parquet file:
        file_name += "_v2"
        # keep the file name below 120 characters and reserve 22 for the max size + parquet extension
        if len(file_name) > 96:
            # sha224 length: 56
            file_name = file_name[:39] + "-" + hashlib.sha224(file_name.encode()).hexdigest()
        file_name += "_max_" + str(config.n_max_entities)
        # add warning for nb max entities to large (>999 999 999 999) because of file name
        return file_name + "." + db_format

    def get_entity_type_string_name(self, entity: pyalex.api.BaseOpenAlex | None = None) -> str:
        """
        Gets the entity type in the format of a string.

        :param entity: The entity type. If nothing is provided, the instance entity type will be used. Default is None.
        :type entity: pyalex.api.BaseOpenAlex | None
        :return: The entity type name in a string
        :rtype: str
        """
        if entity is None:
            entity = self.EntityOpenAlex
        return str(entity).removeprefix("<class 'pyalex.api.").removesuffix("'>").lower()

    def get_entity_type_from_id(self, entity: str | None = None) -> pyalex.api.BaseOpenAlex:
        """
        Gets the entity type from the entity id string.

        :param entity: The entity id. If nothing is provided, the instance entity id will be used. Default is None.
        :type entity: str | None
        :return: The entity type
        :rtype: pyalex.api.BaseOpenAlex
        """
        if entity is None:
            entity = self.entity_from_id
        return get_entity_type_from_id(entity)

    def get_name_of_entity(self, entity: str | None = None) -> str:
        """
        Gets the name of entity from its id.

        :param entity: The entity id, if not provided, use the one from the instance. Default is None.
        :type entity: str
        :return: The name of the entity.
        :rtype: str
        """
        if entity is None:
            entity = self.entity_from_id
            if entity is None:
                raise ValueError("No entity id provided, can't get the name of the entity")
        return get_name_of_entity(entity)

    def get_info_about_entity(self,
                              entity: str | None = None,
                              infos: list[str] | None = None,
                              return_as_pd_series: bool = True,
                              ) -> dict | pd.Series:
        """
        Get information about the entity (eg. name, publication_date...). If no entity is provided, the entity_from_id will be used.

        :param entity: The entity id, if not provided, use the one from the instance. Default is None.
        :type entity: str | None
        :param infos: The information fields to get. Default is None, which will get ["display_name"].
        :type infos: list[str] | None
        :param return_as_pd_series: True to return the results as a Pandas Series. Otherwise, a dictionary is returned. Default is True.
        :type return_as_pd_series: bool
        :return:
        :rtype:
        """
        if entity is None:
            entity = self.entity_from_id
            if entity is None:
                raise ValueError("Can't get the entity info of None")
        return get_info_about_entity(entity, infos=infos, return_as_pd_series=return_as_pd_series)


    def get_multiple_entities_from_id(self, ids: list[str],
                                      ordered: bool = True,
                                      return_dataframe: bool = True) -> pd.DataFrame | list:
        """
        Get multiple entities from their OpenAlex IDs by querying them to the OpenAlex API 100 by 100.

        :param ids: the list of OpenAlex IDs to query
        :type ids: list[str]
        :param ordered: keep the order of the input list in the output list. Default is True.
        :type ordered: bool
        :param return_dataframe: Return a Dataframe. If True, the DataFrame returned will also be stored in self.entities_df and the cache system will be used. If False, a list will be returned. Default is True.
        :type return_dataframe: bool
        :return: the list of entities as pyalex objects (dictionaries) or DataFrame.
        :rtype: pd.DataFrame | list
        """
        res = [] * len(ids)
        i = 0
        with tqdm(total=len(ids), disable=config.disable_tqdm_loading_bar) as pbar:
            # reduce 100 if too big for OpenAlex
            while i + 100 < len(ids):
                res[i:i+100] = self.EntityOpenAlex().filter(ids={'openalex': '|'.join(ids[i:i+100])}).get(per_page=100)
                i += 100
                pbar.update(100)
            res[i:] = self.EntityOpenAlex().filter(ids={'openalex': '|'.join(ids[i:])}).get(per_page=100)
            pbar.update(len(ids) % 100)

        if ordered:
            # sort the res list with the order provided in the list ids
            # create a dictionary with each id as key and the index in the res list as value
            res_ids_index = {entity['id'][21:]: i for i, entity in enumerate(res) if entity is not None}
            # sort based on the index
            res = [res[res_ids_index[entity_id]] if res_ids_index.get(entity_id) is not None else None
                   for entity_id in ids]

        if return_dataframe:
            # similar to the download function, apply the needed formatting (e.g. extracting the abstract for works)
            for i in range(len(res)):
                if res[i] is not None:
                    self.filter_and_format_entity_data_from_api_response(res[i])
            res = self.convert_entities_list_to_df(res)
        return res

def get_entity_type_from_id(entity: str) -> pyalex.api.BaseOpenAlex:
    """
     Gets the entity type from the entity id string.

    :param entity: The entity id, if not provided, use the one from the instance. Default is None.
    :type entity: str | None
    :return: The entity type
    :rtype: pyalex.api.BaseOpenAlex
    """
    match entity[0]:
        case 'W':
            return Works
        case 'A':
            return Authors
        case 'S':
            return Sources
        case 'I':
            return Institutions
        case 'T':
            return Topics
        case 'C':
            return Concepts
        case 'P':
            return Publishers
        case _:
            raise ValueError("Entity id " + entity + " not valid")


def get_name_of_entity(entity: str) -> str:
    """
    Gets the name of the entity from the api.

    :param entity: The entity id.
    :type entity: str
    :return: The name of the entity.
    :rtype: str
    """
    # call the API
    e = get_entity_type_from_id(entity)()[entity]
    return e['display_name']


def extract_authorships_citation_style(authorships: dict) -> str:
    """
    Create a string to cite the authors from an authorships dictionary.

    :param authorships: The authorships dictionary.
    :type authorships: dict
    :return: The citation string.
    :rtype: str
    """
    if len(authorships) == 0:
        res = "Unknown author"
    else:
        res = authorships[0]['author']['display_name']
    if len(authorships) >= 2:
        res += ", " + authorships[1]['author']['display_name']
    if len(authorships) > 2:
        res += " et al."
    return res


def get_info_about_entity(entity: str, infos: list[str] | None = None,
                          return_as_pd_series: bool = True) -> str | pd.Series:
    """

    :param entity: The entity id.
    :type entity: str
    :param infos: The information to get (see OpenAlex API documentation). You can also request "author_citation_style" to get a citation string for the authors. The default is None, which will get ["display_name"].
    :type infos: list[str] | None
    :param return_as_pd_series: True to return the results as a Pandas Series. Otherwise, a dictionary is returned. Default is True.
    :type return_as_pd_series: bool
    :return: The entity information.
    :rtype: str | pd.Series
    """
    if infos is None:
        infos = ["display_name"]
    e = get_entity_type_from_id(entity)()[entity]
    if "author_citation_style" in infos:
        e["author_citation_style"] = extract_authorships_citation_style(e["authorships"])
    res = {key: val for key, val in e.items() if key in infos}
    if return_as_pd_series:
        data = [val for val in res.values()]
        index = [key for key in res]
        res = pd.Series(data=data, index=index)
    return res


def check_if_entity_exists(entity: str) -> bool:
    """
    Check if the entity exists.
    TODO: check status code for existence of entity as other errors can occur.

    :param entity: The entity id.
    :type entity: str
    :return: True if the entity exists.
    :rtype: bool
    """
    # get the name of the entity
    api_path = str(entity).removeprefix("<class 'pyalex.api.").removesuffix("'>").lower() + "s"
    # call the API
    response = requests.get("https://api.openalex.org/" + api_path + "/" + entity)
    if response.status_code == 404:
        return False
    else:
        return True


class WorksData(EntitiesData, Works):
    """
    This class contains specific methods for Works entity data.
    """
    EntityOpenAlex = Works

    def filter_and_format_entity_data_from_api_response(self, entity: dict):
        """
        Filter and format the works data downloaded from the API.
        :param entity: The works data from the API.
        :type entity: dict
        :return: The works datas.
        :rtype: dict
        """

        # Store the abstract as a string (here, we avoid the key "abstract" as PyAlex redefined __getitem__() for it)
        entity['extracted_abstract'] = entity['abstract']
        # as we store the abstract as a string, we can delete its inverted index
        del entity['abstract_inverted_index']

    def add_authorships_citation_style(self):
        """
        Add the author_citation_style column to the DataFrame entities_df
        """
        self.entities_df['author_citation_style'] = self.entities_df['authorships'].apply(
            extract_authorships_citation_style)


    def get_multiple_works_from_doi(self, dois: list[str],
                                    ordered: bool = True,
                                    return_dataframe: bool = True) -> pd.DataFrame | list:
        """
        Get multiple works from their DOI by querying them to the OpenAlex API 60 by 60.

        :param dois: the list of DOIs to query
        :type dois: list[str]
        :param ordered: keep the order of the input list in the output list. Default is True.
        :type ordered: bool
        :param return_dataframe: Return a Dataframe. If True, the DataFrame returned will also be stored in self.entities_df and the cache system will be used. If False, a list will be returned. Default is True.
        :type return_dataframe: bool
        :return: the list of works as pyalex objects (dictionaries) or DataFrame.
        :rtype: pd.DataFrame | list
        """
        res = [] * len(dois)
        i = 0
        with tqdm(total=len(dois), disable=config.disable_tqdm_loading_bar) as pbar:
            # querying more than 60 DOIs causes the HTTP query size being larger than what OpenAlex allows
            while i + 60 < len(dois):
                res[i:i+60] = Works().filter(doi='|'.join(dois[i:i+60])).get(per_page=60)
                i += 60
                pbar.update(60)
            res[i:] = Works().filter(doi='|'.join(dois[i:])).get(per_page=60)
            pbar.update(len(dois) % 60)

        if ordered:
            # sort the res list with the order provided in the list dois
            # create a dictionary with each doi as key and the index in the res list as value
            # we use lower as the doi can be valid with either lower or upper cases
            res_dois_index = {entity['doi'].lower(): i for i, entity in enumerate(res) if entity is not None}
            # sort based on the index
            res = [res[res_dois_index[entity_doi.lower()]] if res_dois_index.get(entity_doi.lower()) is not None else None
                   for entity_doi in dois]

        if return_dataframe:
            # similar to the download function, apply the needed formatting (e.g. extracting the abstract)
            for i in range(len(res)):
                if res[i] is not None:
                    self.filter_and_format_entity_data_from_api_response(res[i])
            res = self.convert_entities_list_to_df(res)
        return res


class AuthorsData(EntitiesData, Authors):
    """
    This class contains specific methods for Authors entity data. Not used for now.
    """
    EntityOpenAlex = Authors


class SourcesData(EntitiesData, Sources):
    """
    This class contains specific methods for Sources entity data. Not used for now.
    """
    EntityOpenAlex = Sources


class InstitutionsData(EntitiesData, Institutions):
    """
    This class contains specific methods for Institutions entity data. Not used for now.
    """
    EntityOpenAlex = Institutions


class ConceptsData(EntitiesData, Concepts):
    """
    This class contains specific methods for Concepts entity data. Not used for now.
    """
    EntityOpenAlex = Concepts


class TopicsData(EntitiesData, Topics):
    """
    This class contains specific methods for Topics entity data. Not used for now.
    """
    EntityOpenAlex = Topics


class PublishersData(EntitiesData, Publishers):
    """
    This class contains specific methods for Publishers entity data. Not used for now.
    """
    EntityOpenAlex = Publishers



# Romain THOMAS 2024

import pandas as pd
import numpy as np
import plotly.express as px
import plotly.graph_objects as go

from pyalex import Concepts, Institutions

# config must NOT be imported from pyalex here as it is already imported via entities_analysis

from openalex_analysis.analysis import *

figure_height = 800


class EntitiesPlot:
    """
    EntitiesPlot class which contains generic methods to do plots of OpenAlex entities.
    """
    def get_figure_entities_of_a_concept_color_country(self,
                                                       concept: str,
                                                       plot_parameters: dict | None = None
                                                       ) -> go.Figure:
        """
        Gets the figure with the entities of a concept, and with the country as color.

        :param concept: The concept id.
        :type concept: str
        :param plot_parameters: The plot parameters.
        :type plot_parameters: dict | None
        :return: The figure.
        :rtype: go.Figure
        """
        if plot_parameters is None:
            plot_parameters = {
                'plot_title': "Plot of the entities related to " + Concepts()[concept]['display_name'] + " studies",
                'x_datas': 'works_count',
                'x_legend': "Number of works",
                'y_datas': concept,
                'y_legend': "Concept score (" + Concepts()[concept]['display_name'] + ")",
            }
        plot_title = plot_parameters['plot_title']
        x_datas = plot_parameters['x_datas']
        x_legend = plot_parameters['x_legend']
        y_datas = plot_parameters['y_datas']
        y_legend = plot_parameters['y_legend']
        color_data = self.getCustomData(concept)[1]
        color_legend = 'Country name'

        fig = px.scatter(
            self.entities_df,
            x=x_datas,
            y=y_datas,
            log_x=True,
            custom_data=self.getCustomData(concept),
            color=color_data,
            # sort the elements and push the None elements at the end
            category_orders={color_data: sorted(self.entities_df[color_data], key=lambda x: (x is None, x))},
            labels={x_datas: x_legend, y_datas: y_legend, color_data: color_legend},
            height=figure_height
        )

        fig.update_traces(hovertemplate="<br>".join(self.getHoverTemplate(concept)))

        fig.update_yaxes(type='linear')

        fig.update_layout(
            title={
                'text': plot_title,
                'x': 0.5,
                'xanchor': 'center',
                'yanchor': 'top'
            },
        )

        return fig


    def get_figure_time_series_element_used_by_entities(self,
                                                        element: str | None = None,
                                                        plot_title: str | None = None,
                                                        x_datas: str = 'year',
                                                        x_legend: str = "Year",
                                                        y_datas: list[str] | None = None,
                                                        ) -> go.Figure:
        """
        Get the figure with the time series usage of a element (eg. reference, concept) by entities.

        :param element: The element id. The default value is None. When the element is None, the first element id in the element_count_df is selected.
        :type element: str | None
        :param plot_title: The plot title. The default value is None to generate an appropriate title.
        :type plot_title: str | None
        :param x_datas: The x datas. The default value is 'year'.
        :type x_datas: str
        :param x_legend: The x legend. The default value is "Year".
        :type x_legend: str
        :param y_datas: The y datas (the entities to plot). The default value is None to use all the entities in the dataframe.
        :type y_datas: list[str] | None
        :return: The figure.
        :rtype: go.Figure
        """
        df = self.element_count_df

        if element is None:
            # take the first element to plot
            element = df.index[0][0]
        element_id = element.strip("https://openalex.org/")
        # if y_legend == None:
        #     y_legend = self.get_name_of_entitie(element.strip("https://openalex.org/"))
        if plot_title is None:
            element_name = self.get_name_of_entity(element_id)
            if len(element_name) > 70:
                element_name = element_name[0:70] + "..."
            plot_title = "Plot of the yearly usage of " + element_id + " (" + element_name + ") by the entities"
        if y_datas is None:
            y_datas = self.count_entities_cols

        df = df[y_datas].loc[element].reset_index()

        df = pd.melt(df, id_vars=x_datas, value_vars=df.columns[1:], var_name='entitie', value_name='nb_used')

        fig = px.line(
            df,
            x=x_datas,
            y='nb_used',
            color='entitie',
            labels={x_datas: x_legend}
        )

        fig.update_layout(
            title={
                'text': plot_title,
                'x': 0.5,
                'xanchor': 'center',
                'yanchor': 'top'
            },
        )

        return fig


    def get_figure_collaborations_with_institutions(self,
                                                    plot_title: str | None = None,
                                                    markers_size_scale: float = 0.3
                                                    ) -> go.Figure:
        """
        Get the figure with the collaborations with institutions.

        :param plot_title: The plot title. The default value is None to generate an appropriate title.
        :type plot_title: str | None
        :param markers_size_scale: The size scale of the markers. The default value is 0.3.
        :type markers_size_scale: float
        :return: The figure
        :rtype: go.Figure
        """
        self.collaborations_with_institutions_df['marker_size'] = (self.collaborations_with_institutions_df['count'] *
                                                                   markers_size_scale)
        if plot_title is None:
            plot_title = "Collaborations through journal articles"
            if self.collaborations_with_institutions_year is not None:
                plot_title += " in "+str(self.collaborations_with_institutions_year)
            plot_title += (
                '<br><sup>Data from OpenAlex. Made with Openalex Analysis'
                '(<a href="https://github.com/romain894/openalex-analysis">'
                'https://github.com/romain894/openalex-analysis</a>)</sup>'
            )
        fig = px.scatter_geo(
            self.collaborations_with_institutions_df,
                lat='lat',
                lon='lon',
                size='count',
                custom_data=['name', 'country', 'count', 'link_to_works'],
                width=1600, height=figure_height,
                color='name_from',
                labels={"name_from": "Entities"},
                title=plot_title
            )
        # add the hover
        hover_template = [
            "%{customdata[0]}",
            "%{customdata[1]}",
            "<a href=\"%{customdata[3]}\">Click to view collaboration paper(s)</a>",
            "%{customdata[2]} co-authored paper(s)",
        ]
        fig.update_traces(
            hovertemplate="<br>".join(hover_template),
            marker=dict(
                opacity=1,
                sizemode='area',
                sizeref=markers_size_scale,
                line=dict(
                    color="blue",
                    width=0
                )
            ),
        )
        fig.update_layout(title_x=0.5, spikedistance=40)
        fig.update_geos(
            visible=False,
            showcountries=True,
            showland=True,
        )
        # add a marker for each institution in the list of entities_from
        for i, entity in enumerate(self.collaborations_with_institutions_df.id_from.unique()):
            if get_entity_type_from_id(entity) == Institutions:
                fig.add_scattergeo(
                    mode="markers",
                    lat=[self.collaborations_with_institutions_entities_from_metadata.at[entity, 'lat']],
                    lon=[self.collaborations_with_institutions_entities_from_metadata.at[entity, 'lon']],
                    visible = True,
                    marker=dict(
                        color=fig.data[i].marker.color,
                        opacity=1,
                        size=15,
                        line=dict(width=2, color="DarkSlateGrey"),
                        symbol="diamond"
                    ),
                    hovertemplate=[self.collaborations_with_institutions_entities_from_metadata.at[entity, 'name']],
                    name="",
                    showlegend=False
                    )
        return fig


class WorksPlot(EntitiesPlot, WorksAnalysis):
    """
    This class contains specific methods for Works plot.
    """
    def getCustomData(self, concept: str) -> list[str]:
        """
        Get the custom data for the plot.

        :param concept: The concept id.
        :type concept: str
        :return: The list of custom data for Works.
        :rtype: list[str]
        """
        return ['display_name', 'country_name', 'institution_name', 'publication_year', 'cited_by_count', concept]


    def getHoverTemplate(self, concept: str) -> list[str]:
        """
        Get the hover template for the plot.

        :param concept: The concept id.
        :type concept: str
        :return: The hover template.
        :rtype: list[str]
        """
        hover_template = [
            "%{customdata[0]}",
            "Country name: %{customdata[1]}",
            "Institution: %{customdata[2]}",
            "Publication year: %{customdata[3]}",
            "Cited by count: %{customdata[4]}",
            "Concept score (" + Concepts()[concept]['display_name'] + "): %{customdata[5]}",
        ]
        return hover_template


    def get_figure_nb_time_used(self, element_type: str) -> go.Figure:
        """
        Gets the figure with the number of time each reference is used in a list of works. Also work with concepts.

        :param element_type: The element type ('reference' or 'concept').
        :type element_type: str
        :return: The figure.
        :rtype: go.Figure
        """
        n_x = 2000  # x resolution of the plot
        references_works_count = self.get_element_count(element_type=element_type)
        x_references_works_count = np.geomspace(1, len(references_works_count), num=n_x, dtype=int)
        y_references_works_count = [references_works_count[x - 1] for x in x_references_works_count]
        fig = px.line(x=x_references_works_count,
                      y=y_references_works_count,
                      log_x=True,
                      labels={'x': 'Element rank', 'y': 'Number of time used'},
                      title="Number of time the same element is used by the works in " + self.get_name_of_entity(),
                      line_shape='hv')
        return fig


    def get_figure_entities_yearly_usage(self,
                                         count_years: list[int],
                                         entity_used_ids: str | list[str],
                                         entity_from_ids: str | list[str] | None = None,
                                         ) -> go.Figure:
        """
        Get the plot bar figure with the yearly usage of an entity (concept, work) in the works from another entity
        (institution, author).

        :param count_years: The years to count.
        :type count_years: list[int]
        :param entity_used_ids: The entities used.
        :type entity_used_ids: str | list[str]
        :param entity_from_ids: The entities which used the entities to count. The default value is None. If None, the entitie_from_id is used.
        :type entity_from_ids: str | list[str] | None
        :return: THe figure.
        :rtype: go.Figure
        """
        df = self.get_df_yearly_usage_of_entities_by_multiples_entities(
            count_years=count_years,
            entity_used_ids=entity_used_ids,
            entity_from_ids=entity_from_ids,
        )
        fig = px.bar(df,
                     x='years',
                     y='usage_count',
                     color='entity_from',
                     pattern_shape='entity_used',
                     barmode='group',
                     height=600,
                     )
        return fig


    def get_figure_entities_yearly_position(self,
                                            count_years: list[int],
                                            entity_used_ids: str,
                                            entity_from_ids: str | list[str] | None = None,
                                            ) -> go.Figure:
        """
        Get the plot figure with the yearly usage of an entity (concept, work) in the works from another entity
        (institution, author).

        :param count_years: The years to count.
        :type count_years: list[int]
        :param entity_used_ids: The entity used.
        :type entity_used_ids: str
        :param entity_from_ids: The entities which used the entities to count. The default value is None. If None, the entitie_from_id is used.
        :type entity_from_ids: str | list[str] | None
        :return: The figure.
        :rtype: go.Figure
        """
        df = self.get_df_yearly_usage_of_entities_by_multiples_entities(
            count_years=count_years,
            entity_used_ids=entity_used_ids,
            entity_from_ids=entity_from_ids,
        )
        fig = px.line(
            df,
            x='works_count',
            y='usage_count',
            text='years',
            color='entity_from',
            line_dash='entity_used',
            height=600,
        )
        fig.update_traces(textposition="bottom right")
        return fig


class AuthorsPlot(EntitiesPlot, AuthorsAnalysis):
    """
    This class contains specific methods for Authors plot. Not used for now.
    """
    pass


class SourcesPlot(EntitiesPlot, SourcesAnalysis):
    """
    This class contains specific methods for Sources plot. Not used for now.
    """
    pass


class InstitutionsPlot(EntitiesPlot, InstitutionsAnalysis):
    """
    This class contains specific methods for Institutions plot.
    """
    def getCustomData(self, concept: str) -> list[str]:
        """
        Get the custom data for the plot.

        :param concept: The concept id.
        :type concept: str
        :return: The list of custom data for Institutions.
        :rtype: list[str]
        """
        return ['display_name', 'geo.country', 'cited_by_count', 'works_cited_by_count_average', concept]


    def getHoverTemplate(self, concept: str) -> list[str]:
        """
        Get the hover template for the plot.

        :param concept: The concept id.
        :type concept: str
        :return: The hover template.
        :rtype: list[str]
        """
        hover_template = [
            "%{customdata[0]}",
            "Country name: %{customdata[1]}",
            #y_legend+": %{y}",
            "Concept score (" + Concepts()[concept]['display_name'] + "): %{customdata[4]}",
            "Cited by count: %{customdata[2]}",
            "Cited by average: %{customdata[3]}"
        ]
        return hover_template


class ConceptsPlot(EntitiesPlot, ConceptsAnalysis):
    """
    This class contains specific methods for Concepts plot. Not used for now.
    """
    pass


class TopicsPlot(EntitiesPlot, TopicsAnalysis):
    """
    This class contains specific methods for Topics plot. Not used for now.
    """
    pass


class PublishersPlot(EntitiesPlot, PublishersAnalysis):
    """
    This class contains specific methods for Publishers plot. Not used for now.
    """
    pass



from openalex_analysis.plot.entities_plot import config
from openalex_analysis.data.entities_data import load_config_from_file

from openalex_analysis.plot.entities_plot import EntitiesPlot
from openalex_analysis.plot.entities_plot import WorksPlot
from openalex_analysis.plot.entities_plot import AuthorsPlot
from openalex_analysis.plot.entities_plot import SourcesPlot
from openalex_analysis.plot.entities_plot import InstitutionsPlot
from openalex_analysis.plot.entities_plot import TopicsPlot
from openalex_analysis.plot.entities_plot import ConceptsPlot
from openalex_analysis.plot.entities_plot import PublishersPlot

from openalex_analysis.plot.entities_plot import get_entity_type_from_id
from openalex_analysis.plot.entities_plot import get_name_of_entity
from openalex_analysis.plot.entities_plot import get_info_about_entity
from openalex_analysis.plot.entities_plot import check_if_entity_exists

__all__ = [
    "config",
    "load_config_from_file",
    "EntitiesPlot",
    "WorksPlot",
    "AuthorsPlot",
    "SourcesPlot",
    "InstitutionsPlot",
    "TopicsPlot",
    "ConceptsPlot",
    "PublishersPlot",
    "get_entity_type_from_id",
    "get_name_of_entity",
    "get_info_about_entity",
    "check_if_entity_exists",
]



from openalex_analysis.analysis.entities_analysis import config
from openalex_analysis.data.entities_data import load_config_from_file

from openalex_analysis.analysis.entities_analysis import EntitiesAnalysis
from openalex_analysis.analysis.entities_analysis import WorksAnalysis
from openalex_analysis.analysis.entities_analysis import AuthorsAnalysis
from openalex_analysis.analysis.entities_analysis import SourcesAnalysis
from openalex_analysis.analysis.entities_analysis import InstitutionsAnalysis
from openalex_analysis.analysis.entities_analysis import TopicsAnalysis
from openalex_analysis.analysis.entities_analysis import ConceptsAnalysis
from openalex_analysis.analysis.entities_analysis import PublishersAnalysis

from openalex_analysis.analysis.entities_analysis import get_entity_type_from_id
from openalex_analysis.analysis.entities_analysis import get_name_of_entity
from openalex_analysis.analysis.entities_analysis import get_info_about_entity
from openalex_analysis.analysis.entities_analysis import check_if_entity_exists


__all__ = [
    "config",
    "load_config_from_file",
    "EntitiesAnalysis",
    "WorksAnalysis",
    "AuthorsAnalysis",
    "SourcesAnalysis",
    "InstitutionsAnalysis",
    "TopicsAnalysis",
    "ConceptsAnalysis",
    "PublishersAnalysis",
    "get_entity_type_from_id",
    "get_name_of_entity",
    "get_info_about_entity",
    "check_if_entity_exists",
]



# Romain THOMAS 2024

import logging
from collections import Counter

from tqdm import tqdm
import pandas as pd

from pyalex import Works, Authors, Institutions, Concepts

# config must NOT be imported from pyalex here as it is already imported via entities_analysis

from openalex_analysis.data import *


class EntitiesAnalysis(EntitiesData):
    """
    This class contains generic methods to analyse entities.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # variables for the collaborations with institutions dataframe:
        self.collaborations_with_institutions_entities_from_metadata = pd.DataFrame() # ids and metadata of the entities
        # for which to look for their collaborations
        self.collaborations_with_institutions_df = pd.DataFrame() # list of the collaborations
        self.collaborations_with_institutions_year = None # years used to calculate the collaborations


    def get_collaborations_with_institutions(self,
                                             entities_from: list[str] | None = None,
                                             institutions_to_exclude: dict[str, list[str]] | None = None,
                                             year: int | str | None = None,
                                             extra_filters_for_entities_from: dict | None = None
                                             ) -> pd.DataFrame:
        """
        Create the collaborations_with_institutions_df DataFrame.

        :param entities_from: The list of entities to use to count the collaborations. If None, the entity_from_id is used.
        :type entities_from: list[str]
        :param institutions_to_exclude: For each entities_from (key), the list of institutions (value) to exclude when counting the collaborations. For example, it is usefull to exclude the parent institution.
        :type institutions_to_exclude: dict[str, list[str]]
        :param year: The years for which to look for the collaborations. Can be an integer or a string. You can provide a range of years as a string (e.g. "2020-2023")
        :type year: int | str | None
        :param extra_filters_for_entities_from: Filters to be used when downloading the datasets of works to extract the collaboration. If you want to use {'publication_year':2023}, you should use the year parameter and not provide extra_filters_for_entities_from. The extra filter won't be used to generate the links on the plot to check the collaboration works
        :type extra_filters_for_entities_from: dict | None
        :return: The collaborations_with_institutions_df DataFrame
        :rtype: pd.DataFrame
        """
        if entities_from is None:
            if self.entity_from_id is None:
                raise ValueError("You must either provide the entities_from or the entity_from_id when instantiating the object.")
            entities_from = [self.entity_from_id]
        if institutions_to_exclude is None:
            institutions_to_exclude = {}
        if extra_filters_for_entities_from is None:
            extra_filters_for_entities_from = {}
        if year is not None:
            extra_filters_for_entities_from['publication_year'] = year
        self.collaborations_with_institutions_year = year

        # get entities_from metadata
        self.collaborations_with_institutions_entities_from_metadata = [pd.DataFrame()] * len(entities_from)
        for i, entity_id in tqdm(enumerate(entities_from),
                                      total=len(self.collaborations_with_institutions_entities_from_metadata),
                                      desc="Getting entities_from metadata"):
            if get_entity_type_from_id(entity_id) == Institutions:
                inst_obj = Institutions()[entity_id]
                self.collaborations_with_institutions_entities_from_metadata[i] = {'id': inst_obj['id'][21:],
                                                'name': inst_obj['display_name'],
                                                'lat': inst_obj['geo']['latitude'],
                                                'lon': inst_obj['geo']['longitude'],
                                                'country': inst_obj['geo']['country'],
                                                }
            elif get_entity_type_from_id(entity_id) == Authors:
                inst_obj = Authors()[entity_id]
                self.collaborations_with_institutions_entities_from_metadata[i] = {'id': inst_obj['id'][21:],
                                                'name': inst_obj['display_name'],
                                                }
            else:
                raise ValueError("The entity type provided is not valid (only Institutions and Authors are supported).")
        self.collaborations_with_institutions_entities_from_metadata = pd.DataFrame(
            self.collaborations_with_institutions_entities_from_metadata).set_index('id')

        self.collaborations_with_institutions_df = [pd.DataFrame()] * len(entities_from)
        for i, institution_from in enumerate(entities_from):
            log_oa.info(f"Processing {institution_from} "
                        f"({self.collaborations_with_institutions_entities_from_metadata.at[institution_from, 'name']})"
                        f"...")
            # get the institutions to exclude for this institution
            institutions_to_exclude_i = institutions_to_exclude.get(institution_from)
            # if there is no institution to exclude, we only exclude the institution_from
            if institutions_to_exclude_i is None:
                institutions_to_exclude_i = [institution_from]
            else:
                institutions_to_exclude_i.append(institution_from)
            log_oa.info(f"Excluding {len(institutions_to_exclude_i)} institution: {institutions_to_exclude_i}")
            # add the https://openalex.org/ at the beggining of each id
            institutions_to_exclude_i = ["https://openalex.org/" + institution for institution in institutions_to_exclude_i]

            if extra_filters_for_entities_from != {}:
                works = WorksAnalysis(institution_from, extra_filters = extra_filters_for_entities_from)
            else:
                works = WorksAnalysis(institution_from)
            # get the list of institutions who collaborated per work:
            collaborations_per_work = [
                list(set([institution['id'] for author in work for institution in author['institutions']]))
                for work in works.entities_df['authorships'].to_list()
            ]
            # list of the institutions we collaborated with
            institutions_collaborations = set(list(
                [institution for institutions in collaborations_per_work
                 for institution in institutions if institution not in institutions_to_exclude_i]
            ))
            log_oa.info(f"{len(institutions_collaborations)} unique institutions with which "
                        f"{self.collaborations_with_institutions_entities_from_metadata.at[institution_from, 'name']} "
                        f"collaborated")
            # remove the https://openalex.org/ at the beginning
            institutions_collaborations = [institution_id[21:] for institution_id in institutions_collaborations]
            # count the number of collaboration per institutions:
            # collaborations_per_work contains the institutions we collaborated per work, so we
            # can count on how many works we collaborated with each institution
            institutions_count_dict = Counter(list(
                [institution for institutions in collaborations_per_work for institution in institutions
                 if institution not in institutions_to_exclude_i]
            ))

            # create dictionaries with the institution id as key and lon, lat and name as item
            institutions_name = [None] * len(institutions_collaborations)
            institutions_id = [None] * len(institutions_collaborations)
            institutions_lat = [None] * len(institutions_collaborations)
            institutions_lon = [None] * len(institutions_collaborations)
            institutions_country = [None] * len(institutions_collaborations)
            institutions_count = [None] * len(institutions_collaborations)
            institutions = InstitutionsAnalysis().get_multiple_entities_from_id(institutions_collaborations,
                                                                                return_dataframe=False)
            for j, institution in enumerate(institutions):
                institutions_name[j] = institution['display_name']
                institutions_id[j] = institution['id'][21:] # remove https://openalex.org/
                institutions_lat[j] = institution['geo']['latitude']
                institutions_lon[j] = institution['geo']['longitude']
                institutions_country[j] = institution['geo']['country']
                institutions_count[j] = institutions_count_dict[institution['id']]

            # store in a dataframe
            self.collaborations_with_institutions_df[i] = pd.DataFrame(
                list(zip(
                    institutions_name,
                    institutions_id,
                    institutions_lat,
                    institutions_lon,
                    institutions_country,
                    [institution_from] * len(institutions_collaborations),
                    [self.collaborations_with_institutions_entities_from_metadata.at[institution_from, 'name']] *
                         len(institutions_collaborations),
                    institutions_count
                )),
                columns = ['name', 'id', 'lat', 'lon', 'country', 'id_from', 'name_from', 'count']
            )

            self.collaborations_with_institutions_df[i] = self.collaborations_with_institutions_df[i].sort_values(
                'count', ascending=False
            )

        self.collaborations_with_institutions_df = pd.concat(
            self.collaborations_with_institutions_df, ignore_index=True
        )

        # add the link to consult the collaborations works
        if year is not None:
            self.collaborations_with_institutions_df['link_to_works'] = (
                f"https://explore.openalex.org/works?filter=authorships.institutions.lineage:"
                f"{self.collaborations_with_institutions_df.id},"
                f"authorships.institutions.lineage:{self.collaborations_with_institutions_df.id_from},"
                f"publication_year:{year}"
            )
        else:
            self.collaborations_with_institutions_df['link_to_works'] = (
                f"https://explore.openalex.org/works?filter=authorships.institutions.lineage:"
                f"{self.collaborations_with_institutions_df.id},"
                f"authorships.institutions.lineage:{self.collaborations_with_institutions_df.id_from}"
            )

        return self.collaborations_with_institutions_df


class WorksAnalysis(EntitiesAnalysis, WorksData):
    """
    This class contains specific methods for Works entity analysis.
    """
    def get_element_count(self, element_type: str, count_years: list[int] | None = None) -> pd.Series:
        """
        Count the number of times each element (for now references or concepts) is used by the works in self.entities_df
        in total or by year (optional).

        :param element_type: The element type ('reference' or 'concept').
        :type element_type: str
        :param count_years: List of years to count the concepts. The default value is None to not count by years.
        :type count_years: list[int]
        :return: The element count.
        :rtype: pd.Series
        """
        def get_works_references_count() -> pd.Series:
            """
            Count the number of times each referenced work is used by the works in self.entities_df.

            :return: The works references count.
            :rtype: pd.Series
            """
            log_oa.info(f"Creating the works references count of {self.get_entity_type_string_name()}...")
            if count_years is None:
                return self.entities_df['referenced_works'].explode().value_counts().convert_dtypes()
            else:
                counts_df_list = [None] * len(count_years)
                for i, year in enumerate(count_years):
                    counts_df_list[i] = self.entities_df[self.entities_df.publication_year == year][
                        'referenced_works'].explode().value_counts().convert_dtypes()
                entities_count = pd.concat(counts_df_list, axis=1, keys=count_years).reset_index().fillna(0)
                entities_count = entities_count.set_index('referenced_works').stack()
                entities_count.name = 'count'
                return entities_count

        def get_works_concepts_count() -> pd.Series:
            """
            Count the number of times each concept is used by the works in self.entities_df.

            :return: The concept count.
            :rtype: pd.Series
            """
            log_oa.info(f"Creating the concept count of {self.get_entity_type_string_name()}...")
            if count_years is None:
                return self.entities_df['concepts'].explode().apply(
                    lambda c: c['id'] if type(c) == dict else None).value_counts().convert_dtypes()
            else:
                counts_df_list = [None] * len(count_years)
                for i, year in enumerate(count_years):
                    counts_df_list[i] = self.entities_df[self.entities_df.publication_year == year][
                        'concepts'].explode().apply(
                        lambda c: c['id'] if type(c) == dict else None).value_counts().convert_dtypes()
                entities_count = pd.concat(counts_df_list, axis=1, keys=count_years).reset_index().fillna(0)
                entities_count = entities_count.set_index('concepts').stack()
                entities_count.name = 'count'
                return entities_count

        match element_type:
            case 'reference':
                return get_works_references_count()
            case 'concept':
                return get_works_concepts_count()
            case _:
                raise ValueError("Can only count for 'references' or 'concept'")


    def create_element_used_count_array(self,
                                        element_type: str,
                                        entities_from: list[dict] | None = None,
                                        count_years: list[int] | None = None
                                        ):
        """
        Creates the element used count array. Count the number of times each element (e.g. references, concepts...) are
        used. You must provide at least 'element_type' ('reference' or 'concept').
        If you only provide 'element_type' the default behavior is to count the number of time the element_type are used
        (e.g. the number of times each reference is used) in the dataset loaded ('entities_df').
        If you provide 'entities_from', the count will be done for the dataset 'entities_df' if it exists and each
        entity in the list 'entities_from'.
        By default, one count is made by entity, but if you provide 'count_years' the count will be done for each given
        year.
        The result is saved in 'element_count_df'.

        :param element_type: The element type ('reference' or 'concept').
        :type element_type: str
        :param entities_from: The extra entities to which to count the concepts.
        :type entities_from: list[dict]
        :param count_years: If given, it will compute the count for each year separately
        :type count_years: list[int]
        """
        self.count_element_type = element_type
        self.count_element_years = count_years
        self.count_entities_cols = []
        match self.count_element_type:
            case 'reference':
                cols_to_load = ['id', 'referenced_works', 'publication_year']
            case 'concept':
                cols_to_load = ['id', 'concepts', 'publication_year']
            case _:
                raise ValueError("Can only count for 'references' or 'concept'")

        if self.entity_from_id is None and entities_from == []:
            raise ValueError(
                "You need either to instance the object with an entity_from_id or to provide entities_from to "
                "create_element_used_count_array()"
            )

        self.element_count_df = pd.DataFrame()
        self.element_count_df.index.name = self.count_element_type + "s"

        self.create_element_count_array_progress_percentage = 0
        self.create_element_count_array_progress_text = "Creating the " + self.count_element_type + "s array..."

        # Create the count array for the first/main entity if previously added to object
        if self.entity_from_id is not None:
            col_name = self.entity_from_id + " " + self.get_name_of_entity()
            self.count_entities_cols.append(col_name)
            if len(self.entities_df.index) == 0:
                self.element_count_df[col_name] = pd.Series().convert_dtypes()
            else:
                self.element_count_df = pd.concat(
                    [self.element_count_df, self.get_element_count(self.count_element_type, count_years=count_years)],
                    axis=1)
                self.element_count_df = self.element_count_df.rename(columns={'count': col_name})

        if count_years is not None:
            for i, entity in enumerate(entities_from):
                self.create_element_count_array_progress_percentage = int(i / len(entities_from) * 100)
                # initialise the WorksAnalysis instance
                works = WorksAnalysis(**entity, load_only_columns=cols_to_load)
                col_name = works.entity_from_id + " " + works.get_name_of_entity()
                self.count_entities_cols.append(col_name)
                # if there is no data in the dataframe, we add a blank column
                if len(works.entities_df.index) == 0:
                    self.element_count_df[col_name] = pd.Series().convert_dtypes()
                else:
                    self.element_count_df = pd.concat(
                        [self.element_count_df,
                         works.get_element_count(self.count_element_type, count_years=count_years)],
                        axis=1)
                    self.element_count_df = self.element_count_df.rename(columns={'count': col_name})
            self.element_count_df.index = self.element_count_df.index.set_names('element', level=0)
            self.element_count_df.index = self.element_count_df.index.set_names('year', level=1)
        else:
            self.element_count_df.index.name = 'element'

        self.create_element_count_array_progress_percentage = 100


    def sort_count_array(self,
                         sort_by: str = 'h_used_all_l_use_main',
                         sort_by_ascending: bool = False
                         ):
        """
        Sort the dataframe with the count array (element_count_df).

        :param sort_by: The key to sort the dataframe. The default value is 'h_used_all_l_use_main'.
        :type sort_by: str
        :param sort_by_ascending: Whenever to sort the dataframe ascending. The default value is False.
        :type sort_by_ascending: bool
        """
        log_oa.info(f"Sorting by {sort_by}")
        if not self.count_element_years:
            # we didn't count per year so we can do a simple sort
            self.element_count_df = self.element_count_df.sort_values(by=sort_by, ascending=sort_by_ascending)
        else:
            sorted_sums = self.element_count_df[sort_by].groupby(level=0).sum().sort_values(ascending=sort_by_ascending)
            self.element_count_df = self.element_count_df.reindex(sorted_sums.index, level=0)


    def add_statistics_to_element_count_array(self,
                                              sort_by: str = 'h_used_all_l_use_main',
                                              sort_by_ascending: bool = False,
                                              ):
        """
        Adds a statistics to the element count array (statistics between the main entity to compare (second column in
        the dataframe) and the sum of the other entities).
        Note: This function is still experimental and might change in future versions.

        :param sort_by: The key to sort the dataframe. The default value is 'h_used_all_l_use_main'.
        :type sort_by: str
        :param sort_by_ascending: Whenever to sort the dataframe ascending. The default value is False.
        :type sort_by_ascending: bool
        """
        if not self.count_element_type in ['reference', 'concept']:
            raise ValueError("Can only count for 'references' or 'concept'")
        # self.create_references_works_count_array_progress_text = "Adding statistics on the references array..."
        if self.element_count_df.empty:
            raise ValueError("Need to create element_count_df before adding statistics")
        # we need at least 2 entities in the dataframe so 2 columns (self.entity_id and entity or 2 entities)
        # (the ref id are the index)
        nb_entities = len(self.element_count_df.columns)
        if nb_entities < 1:
            raise ValueError("Need at least 2 entities in the dataframe to compare entities")
        main_entity_col_id = self.element_count_df.columns.values[0]
        log_oa.info(f"Main entity: {main_entity_col_id}")
        nb_entities = len(self.element_count_df.columns)
        self.element_count_df.fillna(value=0, inplace=True)
        log_oa.info("Computing sum_all_entities...")
        # self.element_count_df['sum_all_entities'] = self.element_count_df.iloc[:, 1:1+nb_entities].sum(axis=1)
        self.element_count_df['sum_all_entities'] = self.element_count_df.sum(axis=1)
        log_oa.info("Computing average_all_entities...")
        self.element_count_df['average_all_entities'] = self.element_count_df['sum_all_entities'] / nb_entities
        log_oa.info("Computing proportion_used_by_main_entity")
        # use sum all entities (include main entity in the sum)
        log_oa.info("fill with NaN values 0 of sum_all_entities to avoid them to be used when ranking (we want to"
                    "ignore these rows as these references aren't used)")
        self.element_count_df['sum_all_entities'] = self.element_count_df['sum_all_entities'].replace(0, None)
        self.element_count_df['proportion_used_by_main_entity'] = self.element_count_df[main_entity_col_id] / \
                                                                  self.element_count_df['sum_all_entities']
        # # we put -1 inplace of NaN values (it's where the sum_all_entities is 0 so the division failed)
        # self.element_count_df.fillna(value=-1, inplace=True)
        log_oa.info("Computing sum_all_entities rank...")
        self.element_count_df['sum_all_entities_rank'] = self.element_count_df['sum_all_entities'].rank(ascending=True,
                                                                                                        pct=True)  # , method = 'dense') # before method = 'average' was used
        log_oa.info("Computing proportion_used_by_main_entity rank...")
        self.element_count_df['proportion_used_by_main_entity_rank'] = self.element_count_df[
            'proportion_used_by_main_entity'].rank(ascending=False,
                                                   pct=True)  # , method = 'dense') # before method = 'average' was used
        log_oa.info("Computing highly used by all entities and low use by main entity")
        self.element_count_df['h_used_all_l_use_main'] = self.element_count_df['sum_all_entities_rank'] * \
                                                         self.element_count_df['proportion_used_by_main_entity_rank']

        self.sort_count_array(sort_by=sort_by, sort_by_ascending=sort_by_ascending)


    def get_authors_count(self,
                          cols: list[str] | None
                          ) -> pd.DataFrame:
        """
        Count the number of times each author appears in entities_df and return the result as a pd.DataFrame.

        :param cols: Columns to return in the DataFrame. Must be existing columns names of authorships. The default value is None which correspond to ['author.id', 'count', 'raw_affiliation_string', 'author.display_name', 'author.orcid'].
        :type cols: list[str]
        :return: The authors count.
        :rtype: pd.DataFrame
        """
        if cols is None:
            cols = ['author.id', 'count', 'raw_affiliation_string', 'author.display_name', 'author.orcid']
        df_authors = pd.json_normalize(self.entities_df['authorships'].explode().to_list())
        authors_count = pd.DataFrame(df_authors.value_counts('author.id'))

        df_authors = df_authors.drop_duplicates('author.id')
        df_authors = df_authors.set_index('author.id')

        authors_count = pd.merge(authors_count, df_authors, how='left', left_index=True, right_index=True).reset_index()

        return authors_count[cols]


    def count_yearly_entity_usage(self, entity: str, count_years: list[int]) -> list[int]:
        """
        Counts the yearly number of time the entity is used in entities_df.

        :param entity: The entity (id) to count.
        :type entity: str
        :param count_years: The years for which we need to count the entity.
        :type count_years: list[int]
        :return: The number of time the entity is used on a yearly basis.
        :rtype: list[int]
        """
        entity_link = "https://openalex.org/" + entity
        count_res = [0] * len(count_years)
        for i, year in enumerate(count_years):
            # get the list of works from the year
            df = self.entities_df.loc[self.entities_df['publication_year'] == year]
            if self.get_entity_type_from_id(entity) == Concepts:
                # get a dataframe with all the concepts used during the year in the column id
                df = pd.json_normalize(df['concepts'].explode().to_list())
            elif self.get_entity_type_from_id(entity) == Works:
                # get a dataframe with all the works used during the year in the column id
                df = pd.DataFrame({'id': df['referenced_works'].explode().dropna()})
            else:
                raise ValueError("Entity type not supported")
            if df.empty:
                count_res[i] = 0
            else:
                # count the concept usage
                count = pd.DataFrame(df.value_counts('id'))
                count_res[i] = count['count'].get(entity_link, 0)
                # count_res[i] = count.loc[concept]['count']
        return count_res


    def count_yearly_works(self, count_years: list[int]) -> list[int]:
        """
        Return the number of works present per year in entities_df.

        :param count_years: The years for which we need to count the works
        :type count_years: list[int]
        :return: Number of works per year.
        :rtype: list[int]
        """
        count_res = [0] * len(count_years)
        for i, year in enumerate(count_years):
            # get the list of works from the year
            df = self.entities_df.loc[self.entities_df['publication_year'] == year]
            if df.empty:
                count_res[i] = 0
            else:
                # get the number of rows
                count_res[i] = len(df.index)
        return count_res


    def get_df_yearly_usage_of_entities(self,
                                        count_years: list[int],
                                        entity_used_ids: str | list[str],
                                        entity_from_legend: str = "Custom dataset"
                                        ) -> pd.DataFrame:
        """
        Gets the dataframe with the yearly usage by works of entity_used_ids.

        :param count_years: The years for which we need to count the entity.
        :type count_years: list[int]
        :param entity_used_ids: The entity ids to count.
        :type entity_used_ids: str | list[str]
        :param entity_from_legend: The legend on the plot for the entity_from dataset. The default value is "Custom dataset". If the default value is unchanged and entitie_from_id was specified, entitie_from_id will be used.
        :type entity_from_legend: str
        :return: The df yearly usage by works.
        :rtype: pd.DataFrame
        """
        if not isinstance(entity_used_ids, list):
            entity_used_ids = [entity_used_ids]
        if entity_from_legend == "Custom dataset" and self.entity_from_id is not None:
            entity_from_legend = self.entity_from_id
        df = pd.DataFrame()
        for entity_used_id in entity_used_ids:
            # count
            usage_count = self.count_yearly_entity_usage(entity_used_id, count_years)
            works_count = self.count_yearly_works(count_years)
            entity_used_id_list = [entity_used_id] * len(count_years)
            entity_from_list = [entity_from_legend] * len(count_years)

            # create the dataframe
            zipped_data = list(zip(count_years,
                                   usage_count,
                                   works_count,
                                   entity_used_id_list,
                                   entity_from_list))

            df_i = pd.DataFrame(zipped_data, columns=['years',
                                                      'usage_count',
                                                      'works_count',
                                                      'entity_used',
                                                      'entity_from',
                                                      ])

            df = pd.concat([df, df_i])

        return df


    def get_df_yearly_usage_of_entities_by_multiples_entities(self,
                                                              count_years: list[int],
                                                              entity_used_ids: str | list[str],
                                                              entity_from_ids: str | list[str] | None = None,
                                                              ) -> pd.DataFrame:
        """
        Gets the dataframe with the yearly usage by works of entity_used_ids, works for multiple entities from.

        :param count_years: The years for which we need to count the entities.
        :type count_years: list[int]
        :param entity_used_ids: The entity ids to count.
        :type entity_used_ids: str | list[str]
        :param entity_from_ids: The entity from identifiers, aka the entities dataset in which we need to count the entity_used_ids. When the default value None is used, the entitie_from_id will be used.
        :type entity_from_ids: str | list[str] | None
        :return: The DataFrame of the yearly usage of entity_used_ids by entity_from_ids
        :rtype: pd.DataFrame
        """
        if entity_from_ids is None:
            entity_from_ids = self.entity_from_id
            if entity_from_ids is None:
                raise ValueError(
                    "entity_from_ids not provided and entities_from_id is None. You must provide either "
                    "entity_from_id to the class or entity_from_ids to the function")

        if not isinstance(entity_from_ids, list):
            entity_from_ids = [entity_from_ids]

        df = pd.DataFrame()

        for entity_from_id in entity_from_ids:
            work_analysis = WorksAnalysis(entity_from_id)
            df = pd.concat([
                df,
                work_analysis.get_df_yearly_usage_of_entities(count_years=count_years,
                                                              entity_used_ids=entity_used_ids,
                                                              )
            ])

        return df


class AuthorsAnalysis(EntitiesAnalysis, AuthorsData):
    """
    This class contains specific methods for Authors entity analysis. Not used for now.
    """
    pass


class SourcesAnalysis(EntitiesAnalysis, SourcesData):
    """
    This class contains specific methods for Sources entity analysis. Not used for now.
    """
    pass


class InstitutionsAnalysis(EntitiesAnalysis, InstitutionsData):
    """
    This class contains specific methods for Institutions entity analysis. Not used for now.
    """
    pass


class ConceptsAnalysis(EntitiesAnalysis, ConceptsData):
    """
    This class contains specific methods for Concepts entity analysis. Not used for now.
    """
    pass


class TopicsAnalysis(EntitiesAnalysis, TopicsData):
    """
    This class contains specific methods for Topics entity analysis. Not used for now.
    """
    pass


class PublishersAnalysis(EntitiesAnalysis, PublishersData):
    """
    This class contains specific methods for Publishers entity analysis. Not used for now.
    """
    pass